# -*- coding: utf-8 -*-

"""
amazoncaptcha.solver
~~~~~~~~~~~~~~~~~~~~

This module contains AmazonCaptcha instance and all the requiries for it.

Attributes:
    MONOWEIGHT (int): The bigger this number - the thicker a monochromed picture
    MAXIMUM_LETTER_LENGTH (int): Maximum letter length by X axis
    MINIMUM_LETTER_LENGTH (int): Minimum letter length by X axis
    SUPPORTED_CONTENT_TYPES (list of str): Used when requesting a captcha url
        to check if Content-Type in the headers is valid

"""

from .utils import cut_the_white, merge_horizontally, find_letter_boxes
from .exceptions import ContentTypeError

from PIL import Image, ImageChops
from io import BytesIO
import warnings
import requests
import json
import zlib
import os

#--------------------------------------------------------------------------------------------------------------

MONOWEIGHT = 1
MAXIMUM_LETTER_LENGTH = 33
MINIMUM_LETTER_LENGTH = 14
SUPPORTED_CONTENT_TYPES = ['image/jpeg']

#--------------------------------------------------------------------------------------------------------------

class AmazonCaptcha(object):

    def __init__(self, img, image_link=None, devmode=False):
        """
        Initializes the AmazonCaptcha instance.

        Args:
            img (str or io.BytesIO): Path to an input image OR an instance
                of BytesIO representing this image.
            image_link (str, optional): Used if `AmazonCaptcha` was created
                using `fromdriver` class method. Defaults to None.
            devmode (bool, optional): If set to True, instead of 'Not solved',
                unrecognised letters will be replaced with dashes.

        """

        self.img = Image.open(img, 'r')
        self._image_link = image_link
        self.devmode = devmode

        self.letters = dict()
        self.result = dict()

        package_directory_path = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))
        self.training_data_folder = os.path.join(package_directory_path, 'training_data')
        self.alphabet = [filename.split('.')[0] for filename in os.listdir(self.training_data_folder)]

    @property
    def image_link(self):
        """
        Image link property is being assigned only if the instance was
        created using `fromdriver` or `fromlink` class methods.

        If you have created an AmazonCaptcha instance using the constructor,
        the property will be equal to None which triggers the warning.

        """

        if not self._image_link:
            warnings.warn("Seems like you are trying to pull out the image link while not having it.", Warning, stacklevel=2)

        return self._image_link

    def _monochrome(self):
        """
        Makes a captcha pure monochrome.

        Literally says: "for each pixel of an image turn codes 0, 1 to a 0,
        while everything in range from 2 to 255 should be replaced with 255".
        *All the numbers stay for color codes.
        """

        self.img = self.img.convert('L')
        self.img = Image.eval(self.img, lambda a: 0 if a <= MONOWEIGHT else 255)

    def _find_letters(self):
        """
        Extracts letters from an image using found letter boxes.

        Populates 'self.letters' with extracted letters being PIL.Image instances.
        """

        letter_boxes = find_letter_boxes(self.img, MAXIMUM_LETTER_LENGTH)
        letters = [self.img.crop((letter_box[0], 0, letter_box[1], self.img.height)) for letter_box in letter_boxes]

        if (len(letters) == 6 and letters[0].width < MINIMUM_LETTER_LENGTH) or (len(letters) != 6 and len(letters) != 7):
            letters = [Image.new('L', (200, 70)) for i in range(6)]

        if len(letters) == 7:
            letters[6] = merge_horizontally(letters[6], letters[0])
            del letters[0]

        letters = [cut_the_white(letter) for letter in letters]
        self.letters = {str(k): v for k, v in zip(range(1, 7), letters)}

    def _save_letters(self):
        """
        Transforms separated letters into pseudo binary.

        Populates 'self.letters' with pseudo binaries.
        """

        for place, letter in self.letters.items():
            letter_data = list(letter.getdata())
            letter_data_string = ''.join(['1' if pix == 0 else '0' for pix in letter_data])

            pseudo_binary = str(zlib.compress(letter_data_string.encode('utf-8')))
            self.letters[place] = pseudo_binary

    def _translate(self):
        """
        Finds patterns to extracted pseudo binary strings from data folder.

        Literally says: "for each pseudo binary scan every stored letter
        pattern and find a match".

        Returns:
            str: a solution if there is one OR
                'Not solved' if devmode set to False OR
                a solution where unrecognised letters will be replaces with dashes

        """

        for place, pseudo_binary in self.letters.items():
            for letter in self.alphabet:

                with open(os.path.join(self.training_data_folder, letter + '.json'), 'r', encoding = 'utf-8') as js:
                    data = json.loads(js.read())

                if pseudo_binary in data:
                    self.result[place] = letter
                    break

            else:
                self.result[place] = '-'

                if not self.devmode:
                    return 'Not solved'

        return ''.join(self.result.values())

    def solve(self, keep_logs=False, logs_path='not-solved-captcha.log'):
        """
        Runs the sequence of solving a captcha.

        Args:
            keep_logs (bool): Not solved captchas will be logged if True.
                Defaults to False.
            logs_path (str): Path to the file where not solved captcha
                links will be stored. Defaults to "not-solved-captcha.log".

        Returns:
            str: Result of the sequence.

        """

        self._monochrome()
        self._find_letters()
        self._save_letters()

        solution = self._translate()

        if solution == 'Not solved' and keep_logs and self.image_link:

            with open(logs_path, 'a', encoding='utf-8') as f:
                f.write(self.image_link + '\n')

        return solution

    @classmethod
    def fromdriver(cls, driver, devmode=False):
        """
        Takes a screenshot from your webdriver, crops the captcha, and stores
        it into bytes array, which is then used to create an AmazonCaptcha instance.

        This also means avoiding any local savings.

        Args:
            driver (selenium.webdriver.*): Webdriver with opened captcha page.
            devmode (bool, optional): If set to True, instead of 'Not solved',
                unrecognised letters will be replaced with dashes.

        Returns:
            AmazonCaptcha: Instance created based on webdriver.

        """

        png = driver.get_screenshot_as_png()
        element = driver.find_element_by_tag_name('img')
        image_link = element.get_attribute('src')

        location = element.location
        size = element.size
        left = location['x']
        top = location['y']
        right = location['x'] + size['width']
        bottom = location['y'] + size['height']

        img = Image.open(BytesIO(png))
        img = img.crop((left, top, right, bottom))

        bytes_array = BytesIO()
        img.save(bytes_array, format='PNG')
        image_bytes_array = BytesIO(bytes_array.getvalue())

        return cls(image_bytes_array, image_link, devmode)

    @classmethod
    def fromlink(cls, image_link, devmode=False):
        """
        Requests the given link and stores the content of the response
        as `io.BytesIO` that is then used to create AmazonCaptcha instance.

        This also means avoiding any local savings.

        Args:
            link (str): Link to Amazon's captcha image.
            devmode (bool, optional): If set to True, instead of 'Not solved',
                unrecognised letters will be replaced with dashes.

        Returns:
            AmazonCaptcha: Instance created based on the image link.

        Raises:
            ContentTypeError: If response headers contain unsupported
                content type.

        """

        response = requests.get(image_link)

        if response.headers['Content-Type'] not in SUPPORTED_CONTENT_TYPES:
            raise ContentTypeError(response.headers['Content-Type'])

        image_bytes_array = BytesIO(response.content)

        return cls(image_bytes_array, image_link, devmode)

#--------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

"""
amazoncaptcha.devtools
~~~~~~~~~~~~~~~~~~~~~~

This module contains the set of amazoncaptcha's devtools.
"""

from .solver import AmazonCaptcha
from .exceptions import NotFolderError
from .__version__ import __version__

from io import BytesIO
import multiprocessing
import requests
import os

#--------------------------------------------------------------------------------------------------------------

class AmazonCaptchaCollector(object):

    def __init__(self, output_folder_path, keep_logs=True, accuracy_test=False):
        """
        Initializes the AmazonCaptchaCollector instance.

        Args:
            output_folder (str): Folder where images or logs should be stored.
            keep_logs (bool, optional): If set to True, unsolved captcha links
                will be stored separately.
            accuracy_test (bool, optional): If set to True, AmazonCaptchaCollector
                will not download images but just solve them and log the results.

        """

        self.output_folder = output_folder_path
        self.keep_logs = keep_logs
        self.accuracy_test = accuracy_test

        if not os.path.exists(self.output_folder):
            os.mkdir(self.output_folder)

        elif not os.path.isdir(self.output_folder):
            raise NotFolderError(self.output_folder)

        self.collector_logs = os.path.join(self.output_folder, f'collector-logs-{__version__.replace(".", "")}.log')
        self.test_results = os.path.join(self.output_folder, 'test-results.log')
        self.not_solved_logs = os.path.join(self.output_folder, 'not-solved-captcha.log')

    def _extract_captcha_link(self, captcha_page):
        """Extracts a captcha link from an html page.

        Args:
            captcha_page (str): A page's html in string format.

        Returns:
            str: Captcha link.

        """

        return captcha_page.text.split('<img src="')[1].split('">')[0]

    def _extract_captcha_id(self, captcha_link):
        """
        Extracts a captcha id from a captcha link.

        Args:
            captcha_link (str): A link to the captcha image.

        Returns:
            str: Captcha ID.

        """

        return ''.join(captcha_link.split('/captcha/')[1].replace('.jpg', '').split('/Captcha_'))

    def get_captcha_image(self):
        """
        Requests the page with Amazon's captcha, gets random captcha.
        Creates AmazonCaptcha instance, stores an original image before solving.

        If it is not an accuracy test, the image will be stored in a specified
        folder with the solution within its name. Otherwise, only the logs
        will be stored, mentioning the captcha link being processed and the result.

        """

        captcha_page = requests.get('https://www.amazon.com/errors/validateCaptcha')
        captcha_link = self._extract_captcha_link(captcha_page)

        response = requests.get(captcha_link)
        captcha = AmazonCaptcha(BytesIO(response.content))
        captcha._image_link = captcha_link
        original_image = captcha.img

        solution = captcha.solve(keep_logs=self.keep_logs, logs_path=self.not_solved_logs)
        log_message = f'{captcha.image_link}::{solution}'

        if solution != 'Not solved' and not self.accuracy_test:
            print(log_message)
            captcha_name = 'dl_' + self._extract_captcha_id(captcha.image_link) + '_' + solution + '.png'
            original_image.save(os.path.join(self.output_folder, captcha_name))

        else:
            print(log_message)
            with open(self.collector_logs, 'a', encoding='utf-8') as f:
                f.write(log_message + '\n')

    def _distribute_collecting(self, milestone):
        """Distribution function for multiprocessing."""

        for step in milestone:
            self.get_captcha_image()

    def start(self, target, processes):
        """
        Starts the process of collecting captchas or conducting a test.

        Args:
            target (int): Number of captchas to be processed.
            processes (int): Number of simultaneous processes.

        """

        goal = list(range(target))
        milestones = [goal[x: x + target // processes] for x in range(0, len(goal), target // processes)]

        jobs = list()
        for j in range(processes):
            p = multiprocessing.Process(target=self._distribute_collecting, args=(milestones[j], ))
            jobs.append(p)
            p.start()

        for proc in jobs:
            proc.join()

        if self.accuracy_test:
            with open(self.collector_logs, 'r', encoding='utf-8') as f:
                output = f.readlines()

            all_captchas = len(output)
            solved_captchas = len([i for i in output if 'Not solved' not in i])
            success_percentage = round((solved_captchas / all_captchas) * 100, 5)
            result = f'::Test::Ver{__version__}::Cap{all_captchas}::Per{success_percentage}::'

            with open(self.test_results, 'w', encoding='utf-8') as f:
                print(result)
                f.write(result)

#--------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

"""
amazoncaptcha.exceptions
~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the set of amazoncaptcha's exceptions.
"""

#--------------------------------------------------------------------------------------------------------------

class ContentTypeError(Exception):
    """
    Requested url, which was supposed to be the url to the captcha image,
    contains unsupported content type within response headers.
    """

    def __init__(self, content_type, message='is not supported as a Content-Type. Cannot extract the image.'):
        self.content_type = content_type
        self.message = message

    def __str__(self):
        return f'"{self.content_type}" {self.message}'

class NotFolderError(Exception):
    """
    Given path, which was supposed to be the path to the folder where
    script can store images, is not a folder.
    """

    def __init__(self, path, message='is not a folder. Cannot store images there.'):
        self.path = path
        self.message = message

    def __str__(self):
        return f'"{self.path}" {self.message}'

#--------------------------------------------------------------------------------------------------------------



__title__ = 'amazoncaptcha'
__description__ = "Pure Python, lightweight, Pillow-based solver for the Amazon text captcha."
__url__ = 'https://github.com/a-maliarov/amazoncaptcha'
__version__ = '0.5.5'
__author__ = 'Anatolii Maliarov'
__author_email__ = 'tly.mov@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020 Anatolii Maliarov'



# -*- coding: utf-8 -*-

"""Solver for Amazon's image captcha.

The motivation behind the creation of this library is taking its start from
the genuinely simple idea: "I don't want to use pytesseract or some other
non-amazon-specific OCR services, nor do I want to install some executables to
just solve a captcha. I desire to get a solution within 1-2 lines of code
without any heavy add-ons. Using a pure Python."

Examples:
    Browsing Amazon using selenium and stuck on captcha? The class method
    below will do all the "dirty" work of extracting an image from the webpage
    for you. Practically, it takes a screenshot from your webdriver, crops the
    captcha, and stores it into bytes array, which is then used to create an
    AmazonCaptcha instance. This also means avoiding any local savings.

        from amazoncaptcha import AmazonCaptcha
        from selenium import webdriver

        driver = webdriver.Chrome() # This is a simplified example
        driver.get('https://www.amazon.com/errors/validateCaptcha')

        captcha = AmazonCaptcha.fromdriver(driver)
        solution = captcha.solve()


    If you are not using selenium or the previous method is not just the case for
    you, it is possible to use a captcha link directly. This class method will
    request the url, check the content type and store the response content into bytes
    array to create an instance of AmazonCaptcha.

        from amazoncaptcha import AmazonCaptcha

        link = 'https://images-na.ssl-images-amazon.com/captcha/usvmgloq/Captcha_kwrrnqwkph.jpg'

        captcha = AmazonCaptcha.fromlink(link)
        solution = captcha.solve()

"""

from .solver import AmazonCaptcha
from .devtools import AmazonCaptchaCollector
from .exceptions import ContentTypeError, NotFolderError

#--------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

import setuptools
import os

#--------------------------------------------------------------------------------------------------------------

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'amazoncaptcha', '__version__.py'), 'r', encoding='utf-8') as f:
    file_data = [i.replace('\n', '').replace('\'', '').split(' = ') for i in f.readlines()]
    about = {k: v for k, v in file_data}

def readme(logo_end_line=14):
    """Extracts the logo from README file before pushing to PyPi."""

    with open('README.md', 'r', encoding='utf-8') as fh:
        long_description = ''.join(fh.readlines()[logo_end_line:])

    return long_description

classifiers = [
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Development Status :: 5 - Production/Stable",
    "Natural Language :: English",
    "Intended Audience :: Developers",
    "Intended Audience :: Education",
    "Intended Audience :: Information Technology",
]

requires = [
    "pillow ~= 9.0.1",
    "requests ~= 2.27.1",
    "selenium >= 3.141,< 4.2"
]

#--------------------------------------------------------------------------------------------------------------

setuptools.setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    packages=['amazoncaptcha'],
    py_modules=['devtools', 'exceptions', 'solver', 'utils'],
    include_package_data=True,
    package_data={'': ['*.json'], 'amazoncaptcha': ['training_data/*.*']},
    classifiers=classifiers,
    long_description=readme(),
    long_description_content_type="text/markdown",
    install_requires=requires,
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    project_urls={
        'Documentation': 'https://amazoncaptcha.readthedocs.io/en/latest/',
        'Source': about['__url__'],
    },
)

#--------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

"""
amazoncaptcha.utils
~~~~~~~~~~~~~~~~~~~

This module contains the set of amazoncaptcha's utilities.
"""

from PIL import Image, ImageChops

#--------------------------------------------------------------------------------------------------------------

def cut_the_white(letter):
    """
    Cuts white spaces/borders to get a clear letter.

    Args:
        letter (PIL.Image): Letter to be processed.

    Returns:
        PIL.Image: The letter without white spaces.

    """

    background = Image.new(letter.mode, letter.size, 255)
    diff = ImageChops.difference(letter, background)
    bbox = diff.getbbox()

    return letter.crop(bbox)

def merge_horizontally(img1, img2):
    """
    Merges two letters horizontally.

    Created in case an image is corrupted and the last letter ends at the
    beginning of the image, causing letter to be unreadable.

    Args:
        img1 (PIL.Image): First letter.
        img2 (PIL.Image): Second letter.

    Returns:
        PIL.Image: Two merged letters.

    """

    merged = Image.new('L', (img1.width + img2.width, img1.height))
    merged.paste(img1, (0, 0))
    merged.paste(img2, (img1.width, 0))

    return merged

def find_letter_boxes(img, maxlength):
    """
    Finds and separates letters from a captcha image.

    Args:
        img (PIL.Image): Monochromed captcha.
        maxlength (int): Maximum letter length by X axis.

    Returns:
        letter_boxes (:obj:`list` of :obj:`tuple`): List with X coords of each letter.

    """

    image_columns = [[img.getpixel((x, y)) for y in range(img.height)] for x in range(img.width)]
    image_code = [1 if 0 in column else 0 for column in image_columns]
    xpoints = [d for d, s in zip(range(len(image_code)), image_code) if s]
    xcoords = [x for x in xpoints if x - 1 not in xpoints or x + 1 not in xpoints]

    if len(xcoords) % 2:
        xcoords.insert(1, xcoords[0])

    letter_boxes = list()
    for s, e in zip(xcoords[0::2], xcoords[1::2]):
        start, end = s, min(e + 1, img.width - 1)

        if end - start <= maxlength:
            letter_boxes.append((start, end))

        else:
            two_letters = {k: v.count(0) for k, v in enumerate(image_columns[start + 5:end - 5])}
            divider = min(two_letters, key=two_letters.get) + 5
            letter_boxes.extend([(start, start + divider), (start + divider + 1, end)])

    return letter_boxes

#--------------------------------------------------------------------------------------------------------------
